@@ -13,7 +13,7 @@ from selenium.webdriver import ActionChains
 from selenium.webdriver.common.keys import Keys
 from slugify import slugify
 
-from klity.features.steps.utils import click, find_button, find_element, find_radio_button, find_checkbox, find_file, find_option, find_link, find_label, find_input, find_recaptcha, find_select, get_value, paste_from_clipboard, send_return
+from klity.features.steps.utils import *
 from klity.klity import TestException
 
 
@@ -68,8 +68,9 @@ def step_impl(context, field, value=""):
     element = find_input(context, get_value(context, field))
     if element is not None:
         element.clear()
+        element.send_keys("")
         if value != "":
-            element.send_keys(get_value(context, value))
+            type_text(context, get_value(context, value))
 
 
 @when(u'je tape ""')
@@ -79,9 +80,7 @@ def step_impl(context, field, value=""):
 def step_impl(context, value=""):
     value = get_value(context, value)
     if value != "":
-        actions = ActionChains(context.browser)
-        actions.send_keys(value)
-        actions.perform()
+        type_text(context, value)
 
 
 @when(u'je clique sur le bouton radio "{value}"')
@@ -152,12 +151,14 @@ def step_impl(context, filename, field):
 def step_impl(context, value):
     value = get_value(context, value)
     start = time.perf_counter()
-    timeout = context.klity.configuration["timeout"]
+    timeout = context.klity.configuration["options"]["timeout"]
     while True:
         if (time.perf_counter() - start) > timeout:
             context.klity.trace(f"=> Timeout after {timeout}s")
             raise TestException(context)
         try:
+            # Deleting old traces
+            context.klity.traces = []
             if find_element(context, value) is not None:
                 return
         except TestException:
@@ -169,7 +170,7 @@ def step_impl(context, value):
 def step_impl(context):
     # Waiting for ReCaptcha
     start = time.perf_counter()
-    timeout = context.klity.configuration["timeout"]
+    timeout = context.klity.configuration["options"]["timeout"]
     while True:
         if (time.perf_counter() - start) > timeout:
             context.klity.trace(f"=> Timeout after {timeout}s.")
@@ -183,6 +184,12 @@ def step_impl(context):
             time.sleep(0.1)
 
 
+@when(u'je soumets le formulaire')
+@when(u'que je soumets le formulaire')
+def step_impl(context):
+    click(context, find_submit(context))
+
+
 ########################################################################################
 # Then
 ########################################################################################
@@ -215,8 +222,8 @@ def step_impl(context, text):
         # espace, espace insécable, tabulation et saut de ligne sont convertis
         "[  \n\t]+", " ", BeautifulSoup(context.browser.page_source, "lxml").text
     )
-    if text not in content:
-        context.klity.trace(f"  Content: {content}")
+    if bytes(text, "utf-8") not in content.encode("utf-8"):
+        context.klity.trace("  Content: %s" % content.encode("utf-8"))
         context.klity.trace("=> Text not found")
         raise TestException(context)
 
@@ -281,13 +288,54 @@ def step_impl(context, value, target=""):
     if target != "":
         href = element.get_attribute("href")
         if target != href:
-            context.klity.trace(f"Le lien ne pointe pas vers {target} mais vers {href}")
+            context.klity.trace(f"Le lien pointe vers '{target}' au lieu de '{href}'")
             raise TestException(context)
 
 
+@then(u'la page contient une image "{value}"')
+def step_impl(context, value):
+    find_image(context, get_value(context,value))
+
+
 @then(u'la variable "{variable}" vaut "{value}"')
 def step_impl(context, variable, value):
-    assert context.klity.variables[variable] == get_value(context,value)
+    value = get_value(context, value)
+    if context.klity.variables[variable] != value:
+        context.klity.trace(f"La variable {variable} vaut '{context.klity.variables[variable]}'")
+        context.klity.trace(f"La valeur attendue est '{value}'")
+        raise TestException(context)
+
+
+@then(u'la variable "{variable_a}" est égale à la variable "{variable_b}"')
+def step_impl(context, variable_a, variable_b):
+    if context.klity.variables[variable_a] != context.klity.variables[variable_b]:
+        context.klity.trace(f"La variable {variable_a} vaut '{context.klity.variables[variable_a]}'")
+        context.klity.trace(f"La variable {variable_b} vaut '{context.klity.variables[variable_b]}'")
+        raise TestException(context)
+
+
+@then(u'la variable "{variable_a}" est différente de la variable "{variable_b}"')
+def step_impl(context, variable_a, variable_b):
+    if context.klity.variables[variable_a] == context.klity.variables[variable_b]:
+        context.klity.trace(f"La variable {variable_a} vaut '{context.klity.variables[variable_a]}'")
+        context.klity.trace(f"La variable {variable_b} vaut '{context.klity.variables[variable_b]}'")
+        raise TestException(context)
+
+
+@then(u'la variable "{variable_a}" est inférieure à la variable "{variable_b}"')
+def step_impl(context, variable_a, variable_b):
+    if context.klity.variables[variable_a] >= context.klity.variables[variable_b]:
+        context.klity.trace(f"La variable {variable_a} vaut '{context.klity.variables[variable_a]}'")
+        context.klity.trace(f"La variable {variable_b} vaut '{context.klity.variables[variable_b]}'")
+        raise TestException(context)
+
+
+@then(u'la variable "{variable_a}" est supérieure à la variable "{variable_b}"')
+def step_impl(context, variable_a, variable_b):
+    if context.klity.variables[variable_a] <= context.klity.variables[variable_b]:
+        context.klity.trace(f"La variable {variable_a} vaut '{context.klity.variables[variable_a]}'")
+        context.klity.trace(f"La variable {variable_b} vaut '{context.klity.variables[variable_b]}'")
+        raise TestException(context)
 
 
 @then(u'le titre du champ "{field}" est "{text}"')
@@ -295,7 +343,10 @@ def step_impl(context, field, text):
     text = get_value(context, text)
     field = get_value(context, field)
     element = find_input(context, field, False)
-    assert element.get_attribute("title") == get_value(context, text)
+    if element.get_attribute("title") != text:
+        context.klity.trace(f"Le titre du champ {field} vaut '{element.get_attribute('title')}'")
+        context.klity.trace(f"La valeur attendue est '{text}'")
+        raise TestException(context)
 
 
 @then(u'le champ "{field}" est vide')
@@ -304,7 +355,16 @@ def step_impl(context, field, text=""):
     text = get_value(context, text)
     field = get_value(context, field)
     element = find_input(context, field, False)
-    assert get_value(context, text) in element.get_attribute("value")
+    if text != "":
+        if text not in element.get_attribute("value"):
+            context.klity.trace(f"Le champ {field} contient '{element.get_attribute('value')}'")
+            context.klity.trace(f"La valeur attendue est '{text}'")
+            raise TestException(context)
+    else:
+        if element.get_attribute("value") != "":
+            context.klity.trace(f"Le champ {field} contient '{element.get_attribute('value')}'")
+            context.klity.trace(f"La valeur attendue est vide")
+            raise TestException(context)
 
 
 @then(u'le champ "{field}" ne contient pas "{text}"')
@@ -312,14 +372,18 @@ def step_impl(context, field, text=""):
     text = get_value(context, text)
     field = get_value(context, field)
     element = find_input(context, field, False)
-    assert get_value(context, text) not in element.get_attribute("value")
+    if text in element.get_attribute("value"):
+        context.klity.trace(f"Le champ {field} contient '{element.get_attribute('value')}'")
+        raise TestException(context)
 
 
 @then(u'le champ "{field}" n\'est pas vide')
 def step_impl(context, field):
     field = get_value(context, field)
     element = find_input(context, field, False)
-    assert element.get_attribute("value") != ""
+    if element.get_attribute("value") == "":
+        context.klity.trace(f"Le champ {field} est vide")
+        raise TestException(context)
 
 
 @then(u'le champ "{field}" contient "{nombre}" caractères')
@@ -328,9 +392,13 @@ def step_impl(context, field, nombre):
     nombre = get_value(context, nombre)
     element = find_input(context, field, False)
     if element.tag_name == "textarea":
-        assert len(element.text) == int(nombre)
+        if len(element.text) != int(nombre):
+            context.klity.trace(f"Le champ {field} possède {len(element.text)} caractères au lieu de {nombre}")
+            raise TestException(context)
     else:
-        assert len(element.get_attribute("value")) == int(nombre)
+        if len(element.get_attribute("value")) != int(nombre):
+            context.klity.trace(f"Le champ {field} possède {element.get_attribute('value')} caractères au lieu de {nombre}")
+            raise TestException(context)
 
 
 @then(u'le champ "{field}" contient moins de "{nombre}" caractères')
@@ -339,9 +407,13 @@ def step_impl(context, field, nombre):
     nombre = get_value(context, nombre)
     element = find_input(context, field, False)
     if element.tag_name == "textarea":
-        assert len(element.text) < int(nombre)
+        if len(element.text) >= int(nombre):
+            context.klity.trace(f"Le champ {field} possède {len(element.text)} caractères, soit plus ou autant que {nombre}")
+            raise TestException(context)
     else:
-        assert len(element.get_attribute("value")) < int(nombre)
+        if len(element.get_attribute("value")) >= int(nombre):
+            context.klity.trace(f"Le champ {field} possède {len(element.get_attribute('value'))} caractères, soit plus ou autant que {nombre}")
+            raise TestException(context)
 
 
 @then(u'le champ "{field}" contient plus de "{nombre}" caractères')
@@ -350,9 +422,13 @@ def step_impl(context, field, nombre):
     nombre = get_value(context, nombre)
     element = find_input(context, field, False)
     if element.tag_name == "textarea":
-        assert len(element.text) > int(nombre)
+        if len(element.text) <= int(nombre):
+            context.klity.trace(f"Le champ {field} possède {len(element.text)} caractères, soit moins ou autant que {nombre}")
+            raise TestException(context)
     else:
-        assert len(element.get_attribute("value")) > int(nombre)
+        if len(element.get_attribute("value")) <= int(nombre):
+            context.klity.trace(f"Le champ {field} possède {len(element.get_attribute('value'))} caractères, soit moins ou autant que {nombre}")
+            raise TestException(context)
 
 
 @then(u'le champ "{field}" contient entre "{nombre_min}" et "{nombre_max}" caractères')
@@ -362,9 +438,13 @@ def step_impl(context, field, nombre_min, nombre_max):
     nombre_max = get_value(context, nombre_max)
     # TODO: This part is to be tested
     if element.tag_name == "textarea":
-        assert int(nombre_min) <= len(element.text) <= int(nombre_max)
+        if not (int(nombre_min) <= len(element.text) <= int(nombre_max)):
+            context.klity.trace(f"Le champ {field} possède {len(element.text)} caractères, donc pas entre {nombre_min} et {nombre_max}")
+            raise TestException(context)
     else:
-        assert int(nombre_min) <= len(element.get_attribute("value")) <= int(nombre_max)
+        if not (int(nombre_min) <= len(element.get_attribute("value")) <= int(nombre_max)):
+            context.klity.trace(f"Le champ {field} possède {len(element.get_attribute('value'))} caractères, donc pas entre {nombre_min} et {nombre_max}")
+            raise TestException(context)
 
 
 @then(u'le champ "{field}" contient l\'option "{value}"')
@@ -373,7 +453,9 @@ def step_impl(context, field, value=""):
     field = get_value(context, field)
     value = get_value(context, value)
     element = find_option(context, field, value)
-    assert element is not None
+    if element is None:
+        context.klity.trace(f"Le champ {field} ne contient pas l'option {value}")
+        raise TestException(context)
 
 
 @then(u'le champ "{field}" ne contient pas l\'option "{value}"')
@@ -385,7 +467,8 @@ def step_impl(context, field, value=""):
     try:
         element = find_option(context, field, value)
         if element is not None:
-            assert False, 'l\'option "%s" du champ "%s" a été trouvée' % (value, field,)
+            context.klity.trace(f"Le champ {field} contient l'option {value}")
+            raise TestException(context)
     except:
         assert True
 
@@ -418,13 +501,21 @@ def step_impl(context, field, value=""):
     except:
         # For radio button
         element = find_radio_button(context, field, value, False)
-    assert element.is_selected()
+    if not element.is_selected():
+        context.klity.trace(f"L'option {value} du champ {field} n'est pas sélectionnée")
+        raise TestException(context)
 
 
 @then(u'la case à cocher "{field}" est cochée')
 def step_impl(context, field):
     field = get_value(context, field)
-    assert find_checkbox(context, field, False).is_selected()
+    element = find_checkbox(context, field, False)
+    if element is None:
+        context.klity.trace(f"La case à cocher {field} n'a pas été trouvée")
+        raise TestException(context)
+    if not element.is_selected():
+        context.klity.trace(f"La case à cocher {field} n'est pas cochée")
+        raise TestException(context)
 
 
 @then(u'la case à cocher "{field}" n\'est pas cochée')
@@ -432,11 +523,12 @@ def step_impl(context, field):
 def step_impl(context, field):
     field = get_value(context, field)
     element = find_checkbox(context, field, False)
-    try:
-        assert not element.is_selected()
-    except AttributeError:
-        # no "is_selected"
-        assert True
+    if element is None:
+        context.klity.trace(f"La case à cocher {field} n'a pas été trouvée")
+        raise TestException(context)
+    if element.is_selected():
+        context.klity.trace(f"La case à cocher {field} est cochée")
+        raise TestException(context)
 
 
 @then(u'je clique sur le bouton "{value}"')
@@ -453,7 +545,22 @@ def step_impl(context, value):
 
 @then(u'un élément contenant "{value}" existe')
 def step_impl(context, value):
-    find_element(context, get_value(context, value))
+    element = find_element(context, get_value(context, value))
+    if element is None:
+        context.klity.trace(f"L'élément {value} n'existe pas ou n'est pas visible.")
+        raise TestException(context)
+
+
+@then(u'un élément contenant "{value}" n\'existe pas')
+def step_impl(context, value):
+    try:
+        element = find_element(context, get_value(context, value))
+    except:
+        pass
+    else:
+        if element is not None:
+            context.klity.trace(f"L'élément {value} existe et est visible.")
+            raise TestException(context)
 
 
 @then(u'le champ "{field}" existe')
@@ -461,21 +568,25 @@ def step_impl(context, field):
     field = get_value(context, field)
     try:
         element = find_input(context, field, False)
-        assert element is not None
     except:
-        assert False, "Champ non trouvé: %s" % field
+        context.klity.trace(f"Le champ {field} n'existe pas.")
+        raise TestException(context)
+    if element is None:
+        context.klity.trace(f"Le champ {field} n'existe pas.")
+        raise TestException(context)
 
 
 @then(u'le champ "{field}" n\'existe pas')
 def step_impl(context, field):
     field = get_value(context, field)
-    element = None
     try:
         element = find_input(context, field, False)
     except:
         pass
-    if element is not None:
-        assert False, "Champ trouvé: %s" % field
+    else:
+        if element is not None:
+            context.klity.trace(f"Le champ {field} existe et est visible.")
+            raise TestException(context)
 
 
 @then(u'le tableau contient "{nombre}" colonnes')
